'''
Created on 04 Oct 2011

@author: Simon Wilshin

Here we encapsulate the code for handling marker sets.

The idea is there will be one generic class which will work under 
most circumstances, then you can plug in others when they are needed.
This is the generic class, in spite of the name, this is probably 
the most complicated of the marker set classes one could write, it is 
simple from the user perspective not in general.
'''

from os.path import join as osjoin
from csv import reader
from re import compile

import conventions as conv
from markerSet import MarkerSetInterface

# Set up the regular expressions that will allow us to go up the leg ticking off joints and bones
reAcc = compile("[FB][RL]Acc")
reGyro = compile("[FB][RL]Gyro")
reMM1Hoof=compile('hoof_dorsal')
reMM2Hoof=compile('hoof_lateral')
reDigit=compile('([a-zA-Z0-9]*(Digit)|(Toe)[a-zA-Z0-9]*)(?i)') # The 'toe' is best thought of as the end of the chain of bones from hip down
reDIP=compile('([a-zA-Z0-9]*(Dip)[a-zA-Z0-9]*)(?i)')
#rePhalanges=compile('([a-zA-Z0-9]*(Phalanges)[a-zA-Z0-9]*)(?i)')
reMTP=compile('([a-zA-Z0-9]*(Fetlock)|(EDL)|(MTP)|(MCP)|(TMP)[a-zA-Z0-9]*)(?i)')
reMetatarsus=compile('([a-zA-Z0-9]*(Metatarsus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Foot)[a-zA-Z0-9]*)(?i)') # Metatarsus is on the foot so put 'foot' labels here.
reTMT=compile('([a-zA-Z0-9]*(TMT)[a-zA-Z0-9]*)(?i)')
reTarsus=compile('([a-zA-Z0-9]*(Tarsus)|(Carpal)|(Tarsal)|(Carpus)[a-zA-Z0-9]*)(?i)') # Heel is roughly the same as the tarsus so put it here
reHock=compile('([a-zA-Z0-9]*(Hock)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Calcaneus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Heel)[a-zA-Z0-9]*)(?i)') # Heel is the Hock so put this here
reTibFib=compile('([a-zA-Z0-9]*(Tibia)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Fibia)[a-zA-Z0-9]*)(?i)')
reStifel=compile('([a-zA-Z0-9]*(Stifle)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Patella)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Knee)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Elbow)[a-zA-Z0-9]*)(?i)') # Where the knee roughly is
reFemur=compile('([a-zA-Z0-9]*(Femur)[a-zA-Z0-9]*)(?i)')
reHip=compile('([a-zA-Z0-9]*(Hip)|(SHO)[a-zA-Z0-9]*)(?i)')
reDistalScapula=compile('([a-zA-Z0-9]*(Disscap)[a-zA-Z0-9]*)(?i)')
reScapula=compile('([a-zA-Z0-9]*(ScapulaRight)|(ScapulaLeft)|(Scapula(!?((_ventral)|(_dorsal))))[a-zA-Z0-9]*)(?i)')
rePelivs=compile('([a-zA-Z0-9]*(Pelvis)|(Proscap)|(Shoudler)|(Shoulder)|(tc)[a-zA-Z0-9]*)(?i)')
legBoneAnatomy=[reAcc, reGyro, reMM1Hoof,reMM2Hoof,reDigit,reDIP,reMTP,reMetatarsus,reTarsus,reHock,reTibFib,reStifel,reFemur,reHip,reScapula,reDistalScapula,rePelivs]
reHoof=compile('(^Toe[a-zA-Z0-9]*)|(^toe[a-zA-Z0-9]*)|(^heel[a-zA-Z0-9]*)|(^dip[a-zA-Z0-9]*)' )

# Now list all markers that will be labelled as head
reHead=compile('([a-zA-Z0-9]*(Head)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Eye)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Ear)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Nose)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Poll)[a-zA-Z0-9]*)(?i)')

# Now list all body markers, not that tail base is a body marker
reBody=compile('([a-zA-Z0-9]*(iliacCrest)|(rtc)|(ltc)|(sacrum)|(Back)|(Belly)|(Body)|(TailBase)|(Spine)|(Thorax)|(Pelvis)|(Illiac)|(Wither(s?))|(Lumbar)|(Cranial)|(Caudal)[a-zA-Z0-9]*)(?i)')

# Now list all tail markers, should not include anythin labelled 'tailbase' so 
# go a bit conservative here
reTail=compile('([a-zA-Z0-9]*(TailTip)[a-zA-Z0-9]*)')

# Define fore and aft and left and right
reLeft=compile('[FB]LAcc|([a-zA-Z0-9]*(Left)|(left)[a-zA-Z0-9]*)|(l_scapula_dorsal)|(l_scapula_ventral)|(l[fh]_[_a-zA-Z0-9]*)|(L[FH]_[_a-zA-Z0-9]*)|([a-z]+L[_a-zA-Z0-9]*)')
reRight=compile('[FB]RAcc|([a-zA-Z0-9]*(Right)|(right)[a-zA-Z0-9]*)|(r_scapula_dorsal)|(r_scapula_ventral)|(r[fh]_[_a-zA-Z0-9]*)|([a-z]+R(?!ear)|(R[FH]_[_a-zA-Z0-9]*)[_a-zA-Z0-9]*)')
reFore=compile('[F][RL]Acc|([a-zA-Z0-9]*(Fore)|(shoudler)|(shoulder)|(Elbow)|(MCP)|(r_scapula_dorsal)|(r_scapula_ventral)|(l_scapula_dorsal)|(l_scapula_ventral)|(Shoulder)|(Scapula)|(scapula)|(carpal)|(Front)[a-zA-Z0-9]*)|(^[a-z]+[RL]F[_a-zA-Z0-9]*$)|(^mcp[LR][_a-zA-Z0-9]*)|(^[rl]f_[a-z]+)|(^[RL]F_[a-z]+)|(^discarpus[LR][_a-zA-Z0-9]*)|(^procarpus[LR][_a-zA-Z0-9]*)|(^elbow[LR][_a-zA-Z0-9]*)|(^sho[LR][_a-zA-Z0-9]*)|(^disscap[LR][_a-zA-Z0-9]*)|(^proscap[LR][_a-zA-Z0-9]*)|([a-zA-Z0-9]*(Carpus)|(Carpal)[a-zA-Z0-9]*)')
reHind=compile('[B][RL]Acc|([a-zA-Z0-9]*(Rear)|(Hind)|(pelvisRight)|(Stifle)|(pelvisLeft)|(MTP)|(RightPelvis)|(LeftPelvis)|(Hip)[a-zA-Z0-9]*)|(^[a-z]+[RL]H)|(^[rl]h_[a-z]+)|(^[RL]H_[a-z]+)|(^mtp[LR])|(^distarsus[LR])|(^protarsus[LR])|(^stifle[LR])|(^hip[LR])|(^tc[LR])|([a-zA-Z0-9]*(Tarsus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(tarsus)|(tarsal)[a-zA-Z0-9]*)')

# Define markers that are labelled but have nothing to do with an animal or robot
reJunkMarkers=compile('([a-zA-Z0-9]*(Camera)|(Obstacle)[a-zA-Z0-9]*)|(marker[0-9])(?i)')

    

# Standardise leg orderings
standardForm={}
standardForm[conv.quadruped]=[[reFore,reLeft],[reHind,reRight],[reHind,reLeft],[reFore,reRight]]
#standardForm[conv.quadruped]=[[reFore,reLeft],[reFore,reRight],[reHind,reRight],[reHind,reLeft]]
standardForm[conv.biped]=[[reLeft],[reRight]]

class ToeID(object):
    '''
    Encapsulates information about where a toe is on a body in general terms.
    This allows us to work out what other markers are on that appendage.
    '''
    def __init__(self,name,legConditions):
        self.name=name
        self.legConditions=legConditions

    def sameLeg(self,marker):
        isSame=True
        for c in self.legConditions:   # Check each condition to be in this leg
            if c.findall(marker)==[]:  # If we are in the same leg this must not be empty
                isSame=False
        return(isSame)

# Set up the functions that allow us to return a suitable ToeID object when we are called on to 
# identify a toe
def identifyMonoPedToe(toe):
    # Pretend we are dealing with a quadruped. 
    cond=[]
    id=""
    if reFore.findall(toe)!=[]:
        id+="Fore"
        cond.append(reFore)
    elif reHind.findall(toe)!=[]:
        id+="Hind"
        cond.append(reHind)
    else:
        raise(LabelException("In a quadruped a foot marker must either be hind or fore!"))
    if reLeft.findall(toe)!=[]:
        id+="Left"
        cond.append(reLeft)
    elif reRight.findall(toe)!=[]:
        id+="Right"
        cond.append(reRight)
    else:
        raise(LabelException("In a quadruped a foot marker must either be left or right!"))
    return(ToeID(id,cond))

def identifyBipedToe(toe):
    # Check if the toe is on the left
    if reLeft.findall(toe)!=[]:
        return(ToeID("Left",[reLeft]))
    elif reRight.findall(toe)!=[]:
        return(ToeID("Right",[reRight]))
    else:
        raise(LabelException("In a quadruped a foot marker must either be left or right!"))    
    
def identifyTripedToe(toe):
    # Check if the toe is on the left, then check if it is fore or aft
    # Pretend that this is a quadruped.
    cond=[]
    id=""
    if reFore.findall(toe)!=[]:
        id+="Fore"
        cond.append(reFore)
    elif reHind.findall(toe)!=[]:
        id+="Hind"
        cond.append(reHind)
    else:
        raise(LabelException("In a quadruped a foot marker must either be hind or fore!"))
    if reLeft.findall(toe)!=[]:
        id+="Left"
        cond.append(reLeft)
    elif reRight.findall(toe)!=[]:
        id+="Right"
        cond.append(reRight)
    else:
        raise(LabelException("In a quadruped a foot marker must either be left or right!"))
    return(ToeID(id,cond))

def identifyQuadrupedToe(toe):
    # Check if the toe is on the left, then check if it is fore or aft
    cond=[]
    id=""
    if reFore.findall(toe)!=[]:
        id+="Fore"
        cond.append(reFore)
    elif reHind.findall(toe)!=[]:
        id+="Hind"
        cond.append(reHind)
    else:
        raise(LabelException("In a quadruped a foot marker must either be hind or fore!"))
    if reLeft.findall(toe)!=[]:
        id+="Left"
        cond.append(reLeft)
    elif reRight.findall(toe)!=[]:
        id+="Right"
        cond.append(reRight)
    else:
        raise(LabelException("In a quadruped a foot marker must either be left or right!"))
    return(ToeID(id,cond))
        
def identifyPentapedToe(toe):
    raise conv.UnwrittenRoutine("Pentaped toe identification routine is unwritten.")

def identifyHexapedToe(toe):
    raise conv.UnwrittenRoutine("Hexaped toe identification routine is unwritten.")

identifyToe=[identifyMonoPedToe,identifyBipedToe,identifyTripedToe,identifyQuadrupedToe,identifyPentapedToe,identifyHexapedToe]

class MarkerSet(MarkerSetInterface):
    '''
    The most generic marker set class. This encapsulates how to catagorise different marker sets.
    It can be replaced when we have those hard to handle marker sets messing things up.
    '''
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        MarkerSetInterface.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self._detectTailMarkers()
        self._checkMarkerClassification()
        self._estimatePolypedalClass()

    def _detectJunkMarkers(self):
        # Some trials are going to have tracked markers we don't want, like cameras, we should ditch these
        self.junkMarkers = [m for m in self.markerNames if reJunkMarkers.findall(m)!=[]]

    def soundOff(self):
        print "All markers ", self.markerNames
        print "Body Markers ", self.bodyMarkers
        print "Head Markers ", self.headMarkers
        print "Tail Markers ", self.tailMarkers
        print "Junk Markers ", self.junkMarkers 
        print "Legs ", self.appendageMarkers
        print "Leg Names ", self.getAppendages()
        
    def getAppendageOfMarker(self,k):
        '''
        FIXME: Should this be in MarkerSetInterface? Appears not.
        Return the appendage that a marker resides on, or none
        '''
        theappend=None
        for append,markers in self.appendageMarkers.iteritems():
            if k in markers:
                theappend=append
        return theappend
        
    def getAppendageAndBodyMarkers(self):
        mks=list()
        mks=self.bodyMarkers
        [mks.extend(k) for k in self.appendageMarkers.itervalues()]
        return(mks)
        
    def _checkMarkerClassification(self):
        '''
        All markers should be classified. We should look at the unclassified 
        markers to check if they are a legs that are just missing a toe and then 
        add them in starting with the lowest tracked marker as the 'toe'.
        
        If after this they are not identified this is problem and 
        we should throw an exception.
        '''
        newAppendages={"":""}
        while newAppendages!={}:
            allIdentifiedMarkers=self.bodyMarkers+self.headMarkers+self.junkMarkers+self.tailMarkers
            for leg in self.appendageMarkers:
                allIdentifiedMarkers+=self.appendageMarkers[leg]
            newAppendages=self._appendageSearch(list(set(self.markerNames)-set(allIdentifiedMarkers)))
            self.appendageMarkers.update(newAppendages)
        if set(allIdentifiedMarkers)!=set(self.markerNames):
            raise(LabelException("Some markers were not identified! Unidentified markers: "+str(set(self.markerNames)-set(allIdentifiedMarkers))))

    def _detectTailMarkers(self):
        '''
        Detects all tail markers in an unordered list
        '''
        self.tailMarkers=[k for k in self.markerNames if reTail.findall(k)!=[]]
    
    def _detectBodyMarkers(self):
        '''
        Detects all body  markers in an unordered list
        '''
        self.bodyMarkers=[k for k in self.markerNames if reBody.findall(k)!=[]]
    
    def _detectHeadMarkers(self):
        '''
        Detects all head markers in an unordered list
        '''
        self.headMarkers=[k for k in self.markerNames if reHead.findall(k)!=[]]
    
    def _appendageSearch(self,markers):
        newAppendages={}
        for p in legBoneAnatomy:
            toes=[m for m in markers if p.findall(m)!=[]] 
            if toes!=[]:
                break
        self.numberOfToes+=len(toes)
        for t in toes:
            toeID=identifyToe[self.numberOfToes-1](t) # Call the method for n-1 legs
            legCohabitants=[m for m in markers if toeID.sameLeg(m)==True]
            newAppendages[toeID.name]=[]
            for p in legBoneAnatomy:
                for m in legCohabitants:
                    if p.findall(m)!=[] and [foundMarker for foundMarker in newAppendages[toeID.name] if foundMarker==m] == []:
                        newAppendages[toeID.name].append(m)
        return newAppendages
    
    def _detectAppendages(self):
        '''
        Detect how many legs / fins there are and label them
        '''
        self.numberOfToes=0
        self.appendageMarkers=self._appendageSearch(self.markerNames)
            
    def getNumberOfDetectedLegs(self):
        return(len(self.getAppendages()))

    def _estimatePolypedalClass(self):
        '''
        Detects feet and classifies the animal by the number of feet.
        Does this by counting foot / toe labels. Name the labels sensibly.
        '''
        # By default we don't know the polypedal class. Once we work out 
        # what it is by counting feet, override this with the actual class.
        if not self.__dict__.has_key("polypedalClass"):
            numberOfAppendages=self.getNumberOfDetectedLegs()
            self.polypedalClass=conv.unknown
            if numberOfAppendages==0:
                print "Animal has no appendages, this is likely to be an error"
                self.appendageNames = []
            if numberOfAppendages==2:
                self.appendageNames = ["Left","Right"]
                self.polypedalClass=conv.biped
            if numberOfAppendages==4:
                self.appendageNames = ["ForeLeft","HindRight","HindLeft","ForeRight"]
                self.polypedalClass=conv.quadruped
            if numberOfAppendages==6:
                self.appendageNames = ["ForeLeft","HindRight","HindLeft","HindRight","MidLeft","MidRight"]
                self.polypedalClass=conv.hexaped
            if numberOfAppendages==8:
                self.appendageNames = ["ForeLeft","HindRight","HindLeft","HindRight","MidForeLeft","MidHindRight","MidHindLeft","MidHindRight"]
                self.polypedalClass=conv.octoped
        return(self.polypedalClass)


def getHorseLabels():
    nreader = reader(open(osjoin("..","sampleData","labelConventions","horseLabels.csv"), "r"))
    horseLabels = list()
    for row in nreader:    
        horseLabels.append(row[0])
    return(horseLabels)

def getDogLabels():
    nreader = reader(open(osjoin("..","sampleData","labelConventions","dogLabels.csv"), "r"))
    dogLabels = list()
    for row in nreader:    
        dogLabels.append(row[0])
    return(dogLabels)
    
def testHorseLabels():
    horseLabels = getHorseLabels()
    MarkerSet(horseLabels).soundOff()
    MarkerSet(horseLabels).estimatePolypedalClass()    

def testDogLabels():
    dogLabels = getDogLabels()
    MarkerSet(dogLabels).soundOff()
    MarkerSet(dogLabels).estimatePolypedalClass()  
    
def testSampleMarkerSet():
    sampleLabels=[
                   "LeftFoot",
                   "RightFoot",
                   "Head",
                   "Nose",
                   "LeftEye",
                   "RightEye",
                   "Body",
                   "ShoulderLeft",
                   "ShoulderRight",               
                   "RightKnee"
                 ]
    MarkerSet(sampleLabels).soundOff()
    MarkerSet(sampleLabels).estimatePolypedalClass()

def testRegex(inLabels):
    print "Testing with these markers:"
    print "HOOF:"
    print [k for k in inLabels if not reHoof.findall(k)==[]]    

if __name__=='__main__':
    print "Processing sample marker sets"
    testSampleMarkerSet()
    testHorseLabels()
    testDogLabels()
