'''    
Nested Gaussian Process interpolation 
for oscillatory data with slow varying 
component.
Author: Simon Wilshin
Email: swilshin@rvc.ac.uk
Date: April 2013

TODO: Give this thing the markerset as a target?, so it can find limb + body markers. loop over appendages in marker set take all markers from set of appendages markers that contains this one. okey just use self.parent.resolve('')
TODO: Construct the phase required for each limb + body on first access, use mess from email
TODO: Make the actual interpolation of a time series work on gaps

TODO: Planning with Simon:
  - s=getNanRegions(timeseries) -> returns enumerated list of slices
  - p=getNeighboringPoints(listofslices) -> (rersultof getnanregions) return list of list of indices of closest points
  - ph=phaseEst(listofcoordinates) -> e.g. flatten list of all on limb, see below
    - list of def basePhaseEst(x)
      -returns phase
      - x is Nx(M*3) (numpy array) 
  - nGpMoCap -> just shuffling points and limbs, no interpolation
    - create nested gp int object inside interp, throw away when done (saving intepred marker)
    - have phase est from limb/body use that
    - inside interpolate:
      - loop over x,y,z
        - loop over nanregions
          - use time, phase, coordinate data (500 pts), to _init_ (non train mode)
            - use pred to get out the fill.
              - fill in diction with ngpint keyvalue.
  - nestGPinterpolator
'''

# Because of the low level nature of 
# the maipulations we are going to do here 
# it is easier to just use the gp functions 
# from Nuemann's GPR package directly.
from GPR.gpr import gp_pred, gp_train

# Will we need any of the below in this new version?
import scipy.cluster.hierarchy as hrchy
import numpy as np

from mDInterpret import loadMetaData
from motionCaptureData import MotionCaptureData
from utils import ensure_dir, fillNans
import pylab as pl
from os.path import join as osjoin
import experimentTree

class nestedGPIntMoCapData(MotionCaptureData):
    '''
    Wraps the MotionCaptureData in methods that use a nested gaussian process
    to fill in blanks.
    
    Performs nested gaussian process 
    interpolation of data which can be 
    broken in to two components, a slow 
    varying components which is modelled
    using a gaussian process and a 
    faster varying oscillatory component 
    of known phase which is modelled using 
    a second gaussian process derived after 
    the first.
    This requires us to know the phase of 
    the oscillation and the time of both 
    a training data set and the set to be 
    interpolated.
    Makes use of the existing wrapper for 
    gaussian the Neumann GP library.
    '''
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
    #def __init__(self,X0, Y0, M,sigmaSlow,sigmaFast,snoise,sfac=100.0,train=False):
        '''
        Requires the training data set for 
        the Guassian processes.
        X0 and Y0 are the original data from which the GP is built.
        M is ?
        sigmaSlow is the scale of the slow varying component of the signal (usually time, in sec).
          - For dog gait trans data, 1 s works well.
        sigmaFast is the scale of the fast varying component of the signal (usually phase in radians)
          - For dog gait trans data, 0.1 rad works well
        snoise is the noise scale.
          - For dog gait trans data, 0.1 works well
        sfac is the amount to supress the scale of the slow varying component in during the fast varying 
        interpolation.
        train determines if the fast varying hyperparameters are trained for before running.
    shape of X,Y,M; x=time?; what to use phase or pred?
    
    SHOULD DO THIS BY GAP WITHING THE TRIAL. DOH.
      - Find middle of the gap.
      - Find nearest 500 points. (use distnace ranked: so grabs 500 either side, or 500 to one side if it's an edge gap. Will it work if there's another gap? I suppose you fill it and move forward? Or not fill and move forward.
      
    INPUTS: 
            X0: NxD array; N=data points; D=dimensionality of input space. 1 col of X is time, other is phase
              - For dog: take all limb markers, vstack to Tx3N, subtract with spline X; Y, nan=zero, PCA on that. First Princomp should be the nice wiggly thing. Could diff that.
            Y0: Nx1 
            M: scalars, 1.0 
            sigmaSlow: 
            sigmaFast: 
            snoise: 
            sfac: 
            train: 
        '''
        MotionCaptureData.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        loadMetaData("M",self,self.metaData,1.0)
        loadMetaData("sigmaSlow",self,self.metaData,1.0)
        loadMetaData("sigmaFast",self,self.metaData,0.1)
        loadMetaData("snoise",self,self.metaData,0.1)
        loadMetaData("sfac",self,self.metaData,100)
        loadMetaData("ngptrain",self,self.metaData,False)
        
    def keys(self):
        return([k for k in self.data.keys() if k[-7:]!="_ngpint"])
        
    def has_key(self,k):
        return(k in self.keys())
  
    def phaseModel(self,Xs):
        if Xs.ndim==1:
            Xs = np.array([Xs]).T
        y0 = gp_pred(
                      self.logThetaFast,
                      self.covFcirc,
                      np.array([self.X0[:,1]]).T,
                      self.Y0,
                      Xs
                    )[0]
        return(y0)
  
    def lowPassFilt(self,Xs):
        if Xs.ndim==1:
          Xs = np.array([Xs]).T
        y0 = gp_pred(
                      self.logThetaSlow,
                      self.covF,
                      np.array([self.X0[:,0]]).T,
                      self.Y0,
                      Xs
                    )[0]
        return(y0)    
  
    def highPassFilt(self,Xs):
        y1 = gp_pred(
                      self.logThetaComb,
                      self.covFARD,
                      self.X0,
                      self.Y1,
                      Xs
                    )[0]
        return(y1)
  
    def pred(self,Xs):
        y0 = self.lowPassFilt(Xs[:,0])
        y1 = self.highPassFilt(Xs)
        return(y0+y1) 

    def train(self):
        '''
        Find suitable filled training data and use it to train gp interpolator.
        
        FIXME: This will use the first experimentTree branch with "Trial" in the iD 
        that contains a toe marker with zero nans across THE ENTIRE EXPERIMENT.
        Should also get subject name maker sure it's within subject.
        Problem is that this mocap object could be at subject (phase train) or trial (trial) levels.
        # First, must train the GP. Let's train on a subject specific, completely filled limbed markers.
        # Find this subject name.
        # Find the experimentTree object.
        # Get nans in everything!
        Use completely filled toe marker from this subject across searching all gait transition trials.
        '''
        
        # Ideally, use callMethod on experiment to return fill rate of all limb markers from all objects. Rank by fill rate, choose best. If all ngp objects do this, should train on same time series.
        # Hrm. Or, more simply, search this mocap object for a toe marker with no nans.
        # Failing that, try all gait transition mocaps for this subject for a toe marker with no nans?
        # Find the root of the tree:
        done=False
        obj=self
        while not done:
            if not isinstance(obj,experimentTree.ExperimentTree):
                obj=obj.parent
            else:
                done=True
        nd=obj.callMethod('nansByMarker')
        # Argh, for some reason this has an ugly tree structure. Ok it's a tree with shape of experiment. At bottom is object where callmethod worked paired with dictionnary of marker results.
        # Well, assume structure always the same, as we are starting at root of tree.
        # nd[0][1] is list at level of subject. Find all of these with 'Trial' in iD:
        found=False
        for item in nd[0][1]:
            if 'Trial' in item[0] and not found: 
                for marker in ['toeForeLeft','toeForeRight','toeRearLeft','toeRearRight']:
                    if item[1][0][1][marker]==0 and not found:
                        found=True
                        # store the mocap objects id
                        mcid=item[1][0][0]
                        # Store the trial id:
                        tid=item[0]
                        print "nGPIntObject: found %s with object iD %s with zero nans in %s marker. Using to train" % (tid,mcid,marker)
        if found==False:
            print "ERROR: nGpIntMoCapData COULD NOT FIND completely filled toe marker in any trial. Will fail"
            sys.exit()
        # Now get that mocap object by iD: <- COULD POINT TO THIS OBJECT!
        mc=obj.resolve(obj.children[0].iD).resolve(tid).resolve(mcid)
        
        # Train the GP on mc.time and the valid marker.
        # Keep track of things
        self.X0 = mc.time
        # Access data directly because using __getitem__ would call this object!
        self.Y0 = mc.data[marker]
        
        # Create the first of our gaussian processes
        # and apply this gp to the training data set 
        # and detrend    
        self.covF = ['kernels.covSum', ['kernels.covSEiso','kernels.covNoise']]
        self.logThetaSlow = np.array([np.log(self.sigmaSlow),np.log(self.M),np.log(self.snoise)])
        y0 = self.lowPassFilt(self.X0[:,0])
        self.covFcirc = ['kernels.covSum', ['kernels.covSEcirciso','kernels.covNoise']]
        
        self.Y1 = self.Y0-y0
        # Now that we have detrended our data we 
        # want to train a second GP to get the 
        # local structure
        self.logThetaFast = np.array([np.log(self.sigmaFast),np.log(self.M),np.log(self.snoise)])
        if ngptrain:
            self.logThetaFast = gp_train(
                                          self.logThetaFast,
                                          self.covFcirc,
                                          np.array([self.X0[:,1]]).T,
                                          self.Y1
                                        )
        # Create the second Guassian process for the detrended data. 
        # The idea here is to base most of the estimation on the phase 
        # information so we rescale the slow portion of the process 
        # so that it in effect just ensures continuity rather than 
        # determining the detailed structure.      
        self.logThetaComb = [
                             np.log(exp(self.np.logThetaSlow[0])*self.sfac),
                             self.logThetaFast[0],
                             self.logThetaFast[1],
                             self.logThetaFast[2]
                            ]
        # This set of hyperparamters will be used with 
        # a automatic revelance detection kernel so 
        # we need a new covariance function.
        self.covFARD = ['kernels.covSum', ['kernels.covSEtauandcirc','kernels.covNoise']]

        # This concludes the initialisation of the GP. Oops. This should be separate, but called on first request for an item.
    
    def interpolate(self,k):
        # For the signal k, find the regions that are NaNs.
        sigs = self.data[k]
        nSigs = []
        for s in sigs:        
            nSigs.append(self.pred(s))
        self.data[k+"_ngpint"] = np.vstack(nSigs)
    
    def __getitem__(self,k):
        if not self.data.has_key(k+"_ngpint"):
            # See if user requested marker from body or appendage that we haven't trained for
            # Identify where marker comes from:
            ms=self.parent.resolve('*tMarkerSet')
            if k in ms["BodyMarkers"]:
                # It's body marker, build body phase for body marker NGP training if we don't have it.
                if not self.data.has_key("ngpbodyphase"):
                    if self.metaData["verbosity"]>0:
                        print "  nested GP building body phase for %s" % k
                    self._buildBodyPhase()
                # Whether we built phase now or earlier, now interpolate using it:
                if self.metaData["verbosity"]>0:
                    print "  nested GP interpolating %s" % k
                self.interpolate(k,self.data["ngpbodyphase"])
            else:
                # TODO: make above elseif in appendages.
                # TODO: add limb lookup to markerset FIX VERY SOON
                # Not body marker. Set limb to limb containing this marker
                for limb in ms.appendageNames:
                    if k in ms[limb+"Markers"]:
                        break
                # Build limb phase for NGP training if we don't have it.
                if not self.data.has_key(limb+"_ngpphase"):
                    if self.metaData["verbosity"]>0:
                        print "  nested GP building body phase for %s" % k
                    self._buildLimbPhase(limb)
                # Whether we built phase now or earlier, now interpolate using it:
                if self.metaData["verbosity"]>0:
                    print "  nested GP interpolating %s" % k
                self.interpolate(k,self.data[limb+"_ngpphase"])                    
        return(self.data[k+"_ngpint"])
