'''
Created on 18 Jan 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

Contains information about the nerve blocking condition used on the animal 
(which legs and what 'block' was applied (for if say saline was used).
'''

from os.path import join as osjoin
from yaml import load as yamlload

from baseData import UnsampledBaseData


class NerveBlockData(UnsampledBaseData):
    '''
    Contains information about a single nerve block, which legs were blocked 
    and what they were blocked with.
    '''
    
    def __init__(self, workingDirectory, analysisDirectory, metaData):
        UnsampledBaseData.__init__(self, workingDirectory, analysisDirectory, metaData, self)
        self.blockedLegs=metaData["blockedLegs"]
        self.blockType=metaData["blockType"]
        
def getSampleNerveBlockData():
    workingDir = osjoin("..","sampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','nerveBlockSample.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(NerveBlockData(workingDir,analysisDir,metaData['nerveBlock']))

if __name__=='__main__':
    '''
    Runs a test using the sample data (which is contained wholly in the YAML in this case)
    '''
    print "Processing sample nerve block data"
    sampleNerveBlockData=getSampleNerveBlockData()
    print "Blocked legs: ", sampleNerveBlockData.blockedLegs
    print "Block type: ", sampleNerveBlockData.blockType