from scipy.interpolate import UnivariateSpline
import numpy as np

from mDInterpret import loadMetaData
from motionCaptureData import MotionCaptureData
from utils import ensure_dir
import pylab as pl
from os.path import join as osjoin

class SplineIntMoCapData(MotionCaptureData):
    '''
    Wraps the MotionCaptureData in methods that use a spline interpolation
    to fill in blanks.
    '''
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        MotionCaptureData.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        
    def keys(self):
        return([k for k in self.data.keys() if k[-6:]!="_spint"])
        
    def has_key(self,k):
        return(k in self.keys())
    
    def interpolate(self,k):
        sigs = self.data[k]
        nSigs = list()
        for s in sigs:
            newS = np.array(s)
            # Get the missing times
            missTau = self.time[np.isnan(s)]
            missIdx = np.array(range(self.time.size))[np.isnan(s)]
            t0 = self.time[~np.isnan(s)]
            X = s[~np.isnan(s)]
            s = UnivariateSpline(t0,X)
            nX = s(self.time)
            newS[missIdx] = nX[missIdx]
            nSigs.append(newS)
        self.data[k+"_spint"] = np.vstack(nSigs)
    
    def __getitem__(self,k):
        if not self.data.has_key(k+"_spint"):
            self.interpolate(k)
        return(self.data[k+"_spint"])

