'''
Created on 20 Nov 2010

@author: Simon Wilshin swilshin@rvc.ac.uk

A class which knows about a working directory and an analysis directory, 
that is a directory where working files are stored (such as raw data) and 
a directory where graphs and other output should be written.
'''

from os.path import isdir
from os.path import exists
from os.path import join as osjoin
from os import makedirs

from mDInterpret import getID

class DirectoryError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class DirectoryAware(object):
    '''
    A class which knows about a working directory and an analysis directory, 
    that is a directory where working files are stored (such as raw data) and 
    a directory where graphs and other output should be written.
    '''


    def __init__(self, baseWorkingDirectory, baseAnalysisDirectory, metaData):
        '''
        Constructor
        '''
        if metaData.has_key("workingDir"):
            self.workingDirectory=osjoin(baseWorkingDirectory,str(metaData["workingDir"]))
        elif type(self).__name__=="ExperimentBranch":
            self.workingDirectory=osjoin(baseWorkingDirectory,getID(metaData))
        else:
            self.workingDirectory = baseWorkingDirectory
        if metaData.has_key("analysisDir"):
            self.analysisDirectory=osjoin(baseAnalysisDirectory,str(metaData["analysisDir"]))
        else:
            self.analysisDirectory=osjoin(baseAnalysisDirectory,str(getID(metaData)))
        # Check the working directory exists and 
        # actually is a directory, if not the user 
        # has probably done something stupid like invoke 
        # the script from the wrong location
        if not isdir(self.workingDirectory):
            raise DirectoryError("Working directory " + self.workingDirectory + " does not exist, are you running in the correct directory?")
        # Make sure the analysis directory exists and is a 
        # directory, if not throw an exception
        if not exists(self.analysisDirectory):
            makedirs(self.analysisDirectory)
        elif not isdir(self.analysisDirectory):
            raise DirectoryError("Unable to create analysis directory " + self.analysisDirectory)
            