'''
A class wrapper around the Gaussian process interpolater by Marion Neumann.

author: Simon Wilshin
'''
from numpy import linalg, array, dot, zeros, size, log, diag, pi, eye,mean
#import scipy.linalg
from Tools.general import feval
import GPR.minimize as minimize

class GPInterpolater(object):
    def __init__(self, X, y, covfunc, loghyper, R=None, w=None, train=False):
        self.X = X
        self.y = y
        self.muy = mean(y)
        self.y = y - self.muy
        self.covfunc = covfunc
        self.loghyper = loghyper
        self.R = R
        self.K = None
        self.w = w
        if train:
            self._gpTrain()

    def _gpTrain(self, logtheta=None):
        if logtheta == None:
            logtheta = self.loghyper
        [self.loghyper, fvals, iter] = minimize.run(logtheta, nlml, dnlml, [self.covfunc, self.X, self.y, self.R, self.w], maxnumfuneval=1000)

    def _makeK(self):
        if self.R==None:
            self.K = feval(self.covfunc, self.loghyper, self.X)                     # training covariances
        else:
            self.K = feval(self.covfunc, self.loghyper, self.X, self.R, self.w)     # training covariances
        self.L = linalg.cholesky(self.K)                                            # cholesky factorization of cov (lower triangular matrix)
        self.alpha = solve_chol(self.L.transpose(),self.y)                          # compute inv(K)*y

    def predict(self, Xstar, Rstar=None):
        if self.K==None:
            self._makeK()
        if self.R==None:
            [Kss, Kstar] = feval(self.covfunc, self.loghyper, self.X, Xstar)        # test covariances (Kss = self covariances, Kstar = cov between train and test cases)
        else:
            [Kss, Kstar] = feval(self.covfunc, self.loghyper, self.X, self.R, self.w, Xstar, Rstar)   # test covariances
        mu = dot(Kstar.transpose(),self.alpha)         # predicted means
        v = linalg.solve(self.L, Kstar)                  
        v2=v*v                                     
        sigma = Kss - array([v2.sum(axis=0)]).transpose()  # predicted variances  
        return [mu + self.muy, sigma]
        
def solve_chol(A,B):
    return linalg.solve(A,linalg.solve(A.transpose(),B))


def nlml(loghyper, covfunc, X, y, R=None, w=None):
 
    n = X.shape[0]
    # compute training set covariance matrix
    if R==None:
        K = feval(covfunc, loghyper, X)
    else:
        K = feval(covfunc, loghyper, X, R, w)     
        
    # cholesky factorization of the covariance
    L = linalg.cholesky(K)      # lower triangular matrix
    # compute inv(K)*y
    alpha = solve_chol(L.transpose(),y)
    # compute the negative log marginal likelihood
    return (0.5*dot(y.transpose(),alpha) + (log(diag(L))).sum(axis=0) + 0.5*n*log(2*pi))[0][0] 


def dnlml(loghyper, covfunc, X, y, R=None, w=None):
    out = zeros((loghyper.shape))
    W = get_W(loghyper, covfunc, X, y, R, w)

    if R==None:
        for i in range(0,size(out)):
            out[i] = (W*feval(covfunc, loghyper, X, i)).sum()/2
    else:
        for i in range(0,size(out)):
            out[i] = (W*feval(covfunc, loghyper, X, R, w, i)).sum()/2
 
    return out    
 
        
def get_W(loghyper, covfunc, X, y, R=None, w=None):
    '''Precompute W for convenience.'''
    n = X.shape[0]
    # compute training set covariance matrix
    if R==None:
        K = feval(covfunc, loghyper, X)
    else:
        K = feval(covfunc, loghyper, X, R, w)

    # cholesky factorization of the covariance
    L = linalg.cholesky(K)      # lower triangular matrix
    alpha = solve_chol(L.transpose(),y)
    W = linalg.solve(L.transpose(),linalg.solve(L,eye(n)))-dot(alpha,alpha.transpose())
    return W

