
import numpy as np
import pylab as pl
from numpy import sin,cos
from scipy.integrate import odeint
from scipy import weave
import sys, time

import toroidalGeometry

D,U = np.linalg.eig(toroidalGeometry.g3T)

# See utils/finalPotentialDerivation
def walkV(x,z):
    return(
        -cos(x[1])*sin(x[0])*z/2 +
        cos(x[1])*sin(x[2])*z/2 +
        sin(x[0])*sin(x[1])/2 -
        sin(x[0])*z/2 +
        sin(x[0])/2 +
        sin(x[1])*sin(x[2])/2 +
        sin(x[2])*z/2 -
        sin(x[2])/2
       )

# See utils/finalPotentialDerivation
# However: note that lambda NEGATIVE gives minima at trot. E.g. -0.5
# Note that lambda POSITIVE gives minima at PACE? NO.
#   Lambda positive and <1 (0.5, or 0.16, as in fits) gives minima at pi,pi,pi = PACE
def trotV(x,lda):
    return(
        (lda/2.0)*cos(x[0]) +
        (lda/2.0)*cos(x[2]) +
        (1-lda)*cos(x[1])
       )

def newV(x,vel,z,lda,mu,nu,chi,rho):
    return( (mu+vel*nu)*walkV(x,z) + (chi+vel*rho)*trotV(x,lda) )

# See utils/finalPotentialDerivationConv2.py -- this has lateral walk (0.25,0.5,0.75) as most stable, as should be.
def walkV2(x,z):
    return(
        cos(x[1])*sin(x[0])*z/2 -
        cos(x[1])*sin(x[2])*z/2 +
        sin(x[0])*sin(x[1])/2 +
        sin(x[0])*z/2 -
        sin(x[0])/2 +
        sin(x[1])*sin(x[2])/2 -
        sin(x[2])*z/2 +
        sin(x[2])/2
       )

def newV2(x,vel,z,lda,mu,nu,chi,rho):
    return( (mu+vel*nu)*walkV2(x,z) + (chi+vel*rho)*trotV(x,lda) )

# Should take x, v, pars
def nndV(x,vel,z,lda,mu,nu,chi,rho,asym=0.0,dasym=0.0,hr=0.0):
    '''
    New potential as rederived by Simon Wilshin October 2012
    '''
    # Gradient of walk potential as a function of phase and parameter z
    # This parameter controls shape of well around walk.
    # Create vector u starting with walk potential
    u = np.array([cos(x[0])*(-cos(x[1])*z + sin(x[1]) - z + 1)/2, cos(x[1])*sin(x[0])/2 + cos(x[1])*sin(x[2])/2 + sin(x[0])*sin(x[1])*z/2 - sin(x[1])*sin(x[2])*z/2, cos(x[2])*(cos(x[1])*z + sin(x[1]) + z - 1)/2])
    #u = np.array([cos(x[0])*sin(x[2]) - g0*sin(x[0]) - cos(x[0])*cos(x[1])/2 - cos(x[0])*cos(x[2])/2 - cos(x[1])*sin(x[0])/2 - sin(x[0])*sin(x[2])/2 + cos(x[0])*sin(x[1])*z/2 - cos(x[0])*sin(x[2])*z - cos(x[1])*g0*sin(x[0]), sin(x[1]) + sin(x[0])*sin(x[1])/2 - sin(x[1])*z - cos(x[0])*sin(x[1])/2 - cos(x[2])*sin(x[1])/2 - sin(x[1])*sin(x[2])/2 + cos(x[2])*g0*sin(x[1]) + cos(x[1])*sin(x[0])*z/2 + cos(x[1])*sin(x[2])*z/2 - cos(x[0])*g0*sin(x[1]), cos(x[2])*sin(x[0]) + g0*sin(x[2]) + cos(x[0])*cos(x[2])/2 + cos(x[1])*cos(x[2])/2 + sin(x[0])*sin(x[2])/2 - cos(x[1])*sin(x[2])/2 + cos(x[1])*g0*sin(x[2]) + cos(x[2])*sin(x[1])*z/2 - cos(x[2])*sin(x[0])*z])
    # Multiply the walk potential by scaling factor including velocity dependent factor
    u *= (mu+vel*nu)
    # Compute the trot potential
    b = np.array([
                   -lda/2.0,
                   1.0-lda,                   
                   -lda/2.0
                  ])
    # Multiply the trot potential by its scaling factor and add
    u += -b*np.sin(x)*(chi+vel*rho)
    if False:
        ph = np.dot(toroidalGeometry.g3T,x)
        u[0]-=hr*4.0*sin(2.0*(ph[0]-ph[2]))
        u[2]+=hr*4.0*sin(2.0*(ph[0]-ph[2]))
        u[0]+=asym+dasym*vel
        u[2]-=asym-dasym*vel
    return(u)

def ndV(x,vel,z,g0,lda,mu,nu,chi,rho,asym=0.0,dasym=0.0,hr=0.0):
    b = np.array([
                   -lda/2.0,
                   1.0-lda,                   
                   -lda/2.0
                  ])
    u = np.array([cos(x[0])*sin(x[2]) - g0*sin(x[0]) - cos(x[0])*cos(x[1])/2 - cos(x[0])*cos(x[2])/2 - cos(x[1])*sin(x[0])/2 - sin(x[0])*sin(x[2])/2 + cos(x[0])*sin(x[1])*z/2 - cos(x[0])*sin(x[2])*z - cos(x[1])*g0*sin(x[0]), sin(x[1]) + sin(x[0])*sin(x[1])/2 - sin(x[1])*z - cos(x[0])*sin(x[1])/2 - cos(x[2])*sin(x[1])/2 - sin(x[1])*sin(x[2])/2 + cos(x[2])*g0*sin(x[1]) + cos(x[1])*sin(x[0])*z/2 + cos(x[1])*sin(x[2])*z/2 - cos(x[0])*g0*sin(x[1]), cos(x[2])*sin(x[0]) + g0*sin(x[2]) + cos(x[0])*cos(x[2])/2 + cos(x[1])*cos(x[2])/2 + sin(x[0])*sin(x[2])/2 - cos(x[1])*sin(x[2])/2 + cos(x[1])*g0*sin(x[2]) + cos(x[2])*sin(x[1])*z/2 - cos(x[2])*sin(x[0])*z])
    #u = np.array([cos(x[0]),-sin(x[1]),-cos(x[2])])-y*np.array([sin(2.0*x[0]),-sin(2.0*x[1]),sin(2.0*x[2])])
    u *= (mu+vel*nu)
    u += -b*np.sin(x)*(chi+vel*rho)
    ph = np.dot(toroidalGeometry.g3T,x)
    u[0]-=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    u[2]+=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    u[0]+=asym+dasym*vel
    u[2]-=asym-dasym*vel
    return(u)

def ndVBulk(x,vel,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr):
    b = np.array([
                   -lda/2.0,
                   1.0-lda,                   
                   -lda/2.0
                  ])
    u = np.array([cos(x[:,0])*sin(x[:,2]) - g0*sin(x[:,0]) - cos(x[:,0])*cos(x[:,1])/2 - cos(x[:,0])*cos(x[:,2])/2 - cos(x[:,1])*sin(x[:,0])/2 - sin(x[:,0])*sin(x[:,2])/2 + cos(x[:,0])*sin(x[:,1])*z/2 - cos(x[:,0])*sin(x[:,2])*z - cos(x[:,1])*g0*sin(x[:,0]), sin(x[:,1]) + sin(x[:,0])*sin(x[:,1])/2 - sin(x[:,1])*z - cos(x[:,0])*sin(x[:,1])/2 - cos(x[:,2])*sin(x[:,1])/2 - sin(x[:,1])*sin(x[:,2])/2 + cos(x[:,2])*g0*sin(x[:,1]) + cos(x[:,1])*sin(x[:,0])*z/2 + cos(x[:,1])*sin(x[:,2])*z/2 - cos(x[:,0])*g0*sin(x[:,1]), cos(x[:,2])*sin(x[:,0]) + g0*sin(x[:,2]) + cos(x[:,0])*cos(x[:,2])/2 + cos(x[:,1])*cos(x[:,2])/2 + sin(x[:,0])*sin(x[:,2])/2 - cos(x[:,1])*sin(x[:,2])/2 + cos(x[:,1])*g0*sin(x[:,2]) + cos(x[:,2])*sin(x[:,1])*z/2 - cos(x[:,2])*sin(x[:,0])*z])
    #u = np.array([cos(x[:,0]),-sin(x[:,1]),-cos(x[:,2])])-y*np.array([sin(2.0*x[:,0]),-sin(2.0*x[:,1]),sin(2.0*x[:,2])])
    u *= (mu+vel*nu).transpose()
    u += ((-b*np.sin(x)).transpose()*(chi+vel*rho))
    # This won't work for x = Nx3, which is what u, above expects!
    ph = np.dot(toroidalGeometry.g3T,x)
    u[0]-=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    u[2]+=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    u[:,0]+=asym+dasym*vel.transpose()
    u[:,2]-=asym-dasym*vel.transpose()
    u = u.transpose()
    return(u)

def ndVBulkTerms(x,vel,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr,nx,plot=False):
    '''
    Used for visualisation etc: since this is not called by the optimisation, it must takes the
    x vector as Nx3, rather than 3xN in the optim func call.
    
    BUG (or I don't understant something): ph below expects x=3xN, but u computation expects Nx3.
    
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    u,dVwalk,walkScale,dVtrot,trotScale,hrterm,asymterm = gaitPotentials.ndVBulkTerms(x,v,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr)
    '''
    b = np.array([
                   -lda/2.0,
                   1.0-lda,                   
                   -lda/2.0
                  ])
    # Expecting x=Nx3, will return u=3xN
    u = np.array([cos(x[:,0])*sin(x[:,2]) - g0*sin(x[:,0]) - cos(x[:,0])*cos(x[:,1])/2 - cos(x[:,0])*cos(x[:,2])/2 - cos(x[:,1])*sin(x[:,0])/2 - sin(x[:,0])*sin(x[:,2])/2 + cos(x[:,0])*sin(x[:,1])*z/2 - cos(x[:,0])*sin(x[:,2])*z - cos(x[:,1])*g0*sin(x[:,0]), sin(x[:,1]) + sin(x[:,0])*sin(x[:,1])/2 - sin(x[:,1])*z - cos(x[:,0])*sin(x[:,1])/2 - cos(x[:,2])*sin(x[:,1])/2 - sin(x[:,1])*sin(x[:,2])/2 + cos(x[:,2])*g0*sin(x[:,1]) + cos(x[:,1])*sin(x[:,0])*z/2 + cos(x[:,1])*sin(x[:,2])*z/2 - cos(x[:,0])*g0*sin(x[:,1]), cos(x[:,2])*sin(x[:,0]) + g0*sin(x[:,2]) + cos(x[:,0])*cos(x[:,2])/2 + cos(x[:,1])*cos(x[:,2])/2 + sin(x[:,0])*sin(x[:,2])/2 - cos(x[:,1])*sin(x[:,2])/2 + cos(x[:,1])*g0*sin(x[:,2]) + cos(x[:,2])*sin(x[:,1])*z/2 - cos(x[:,2])*sin(x[:,0])*z])
    #u = np.array([cos(x[:,0]),-sin(x[:,1]),-cos(x[:,2])])-y*np.array([sin(2.0*x[:,0]),-sin(2.0*x[:,1]),sin(2.0*x[:,2])])
    dVwalk = np.copy(u) # 3xN now
    # why transpose? this is just N...
    # tested and it works: when python gets (3xN * N) it multiplies elementwise along the dimension with N
    u *= (mu+vel*nu).transpose()
    walkScale = np.array((mu+vel*nu).transpose())
    uWalk=walkScale*dVwalk
    # this transpose makes it 3xN
    dVtrot = np.array((-b*np.sin(x)).transpose())
    # unimportant for this because python will match dimensions with length N and elementwise multiply
    u += ((-b*np.sin(x)).transpose()*(chi+vel*rho))
    trotScale = np.array((chi+vel*rho).transpose())
    uTrot=trotScale*dVtrot
    # this should crash, if given x as Nx3. so transposing.
    ph = np.dot(toroidalGeometry.g3T,x.T)
    # ph[0].shape is (1447,) -> ph[0] pulls first axis of ph. so these are N.
    u[0]-=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    u[2]+=hr*4.0*sin(2.0*(ph[0]-ph[2]))
    # assembled this way will be 3xN
    hrterm = np.array([-hr*4.0*sin(2.0*(ph[0]-ph[2])),np.zeros(ph[0].shape),hr*4.0*sin(2.0*(ph[0]-ph[2]))])
    # u is 3xN. So fixed these! why transpose v? it is 1D array?!!??
    u[0,:]+=asym+dasym*vel.transpose()
    u[2,:]-=asym-dasym*vel.transpose()
    asymterm=np.array([asym+dasym*vel.transpose(),np.zeros(u[0].shape),-asym-dasym*vel.transpose()])
    # plot flag:
    pars=[u,uWalk,dVwalk,walkScale,uTrot,dVtrot,trotScale,hrterm,asymterm,x,vel]
    parsl=["u","uWalk","dVwalk","walkScale","uTrot","dVtrot","trotScale","hrterm","asymterm","x","vel"]
    if plot:
        pl.figure()
        for j,k in enumerate(pars):
            if j==0:
                ax0=pl.subplot(len(pars),1,j+1)
            else:
                pl.subplot(len(pars),1,j+1,sharex=ax0)
            #print [optimdata[m][k] for m in np.arange(len(optimdata))]
            if k.ndim>1 and j != 9:
                pl.plot(k.T)
                pl.ylabel(parsl[j])
            else:
                pl.plot(k)
                pl.ylabel(parsl[j])
                if j==9:
                    pl.plot(nx % (np.pi*2.0))
            #if j==0:
            #    pl.title('%s first entry at %s' % (subj+fsuff,optimdata[i]["time"]) )
            #pl.show()
            #f.savefig( subj+fsuff[:-4]+"_params.png" )
    # make it Nx3
    u = u.transpose()
    return(u,uWalk,dVwalk,walkScale,uTrot,dVtrot,trotScale,hrterm,asymterm)

def nsimulate(p0,x0,v,sqrtdt, S):
    x = np.zeros((v.size,3))
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    x[0] = x0
    for i in range(v.size-1):
        x[i+1] = x[i]-ndV(x[i],v[i],z,g0,lda,mu,nu,chi,rho,asym,dasym,hr)*(sqrtdt**2.0)+S*sqrtdt*np.dot(U.transpose(),(np.sqrt(D)*np.random.randn(3)))#np.sum(S*dt*np.dot(g3T,random.gauss(0,1)),1)
    return(x)

def nsimulate2(p0,x0,v,sqrtdt, S):
    x = np.zeros((v.size,3))
    # Converged Billy values
    # time	system	notes	z	lda	mu	nu	chi	rho	err1	err2
    # 20130408_223225	crunch		0.110975866	0.847714628	28.54549405	-0.007366714	-1.264644731	0.00388072	0.10677617	0.194385392
    z,lda,mu,nu,chi,rho = p0
    x[0] = x0
    # def nndV(x,vel,z,lda,mu,nu,chi,rho,asym=0.0,dasym=0.0,hr=0.0):
    for i in range(v.size-1):
        x[i+1] = x[i]-nndV(x[i],v[i],z,lda,mu,nu,chi,rho)*(sqrtdt**2.0)+S*sqrtdt*np.dot(U.transpose(),(np.sqrt(D)*np.random.randn(3)))#np.sum(S*dt*np.dot(g3T,random.gauss(0,1)),1)
    return(x)

def betterSimulate(p0,x0,v,sqrtdt,S):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(v.shape[0]-1)*dt,v.shape[0])
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    x = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),v.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym,hr).transpose(), x0, t)
    return(x)

def betterSimulate_nohr(p0,x0,v,sqrtdt,S):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(v.shape[0]-1)*dt,v.shape[0])
    z,g0,lda,mu,nu,chi,rho,asym,dasym = p0
    x = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),v.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym).transpose(), x0, t)
    return(x)   
    
def betterSimulate_noasymhr(p0,x0,v,sqrtdt,S):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(v.shape[0]-1)*dt,v.shape[0])
    z,g0,lda,mu,nu,chi,rho = p0
    x = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),v.shape[0]-1)],z,g0,lda,mu,nu,chi,rho).transpose(), x0, t)
    return(x)

def betterSimulate_r2(p0,x0,v,sqrtdt,S):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(v.shape[0]-1)*dt,v.shape[0])
    z,lda,mu,nu,chi,rho = p0
    x = odeint(lambda nx,ta: -nndV(nx.transpose(),v[min(int(np.floor(ta/dt)),v.shape[0]-1)],z,lda,mu,nu,chi,rho).transpose(), x0, t)
    return(x)

def nlogLike(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    dW = ((x[1:]-x[:-1]+np.pi)%(2.0*np.pi)-np.pi+(sqrtdt**2.0)*ndVBulk(x,v,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr)[:-1,:])/(sqrtdt)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    S2 = np.var(dW)
    dP = (1.0/np.sqrt(2.0 * np.pi*S2))*np.exp(-(dW**2.0)/(2.0*S2))
    LL = np.sum(-np.log(dP))
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LL
    if verbose:
        return(p0,np.sqrt(S2),LL)
    return(LL)

def leastSq(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(x.shape[0]-1)*dt,x.shape[0])
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    try:
        ix = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym,hr).transpose(), x[0].transpose(), t)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        print "Exception in integrator, exiting"
        sys.exit()
    dW = (x-ix+np.pi)%(2.0*np.pi)-np.pi
    # Penalty for large changes
    P = np.sum(((ix[:-1]-ix[1:]+np.pi)%(2.0*np.pi)-np.pi)>np.pi)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    LS = np.sum(dW**2.0)
    LS+=P
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LS
    if verbose:
        return(p0,LS)
    return(LS)

def derivative_faststop(func):
    def impl(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print sys.exc_info()
            sys.exit(1)
    return impl

@derivative_faststop    
def leastSq2_helper(nx,ta,v,dt,x,p0):
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    #time.sleep(0.01)
    # def ndV(x,vel,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr):
    return -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym,hr).transpose()
    
def leastSq2(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(x.shape[0]-1)*dt,x.shape[0])
    z,g0,lda,mu,nu,chi,rho,asym,dasym,hr = p0
    #try:
    #    ix = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym,hr).transpose(), x[0].transpose(), t)
    ix = odeint(leastSq2_helper,x[0].transpose(),t,args=(v,dt,x,p0))
    #except (KeyboardInterrupt, SystemExit):
    #   raise
    #except:
    #    print "Exception in integrator, exiting"
    #    sys.exit()
    dW = (x-ix+np.pi)%(2.0*np.pi)-np.pi
    # Penalty for large changes
    P = np.sum(((ix[:-1]-ix[1:]+np.pi)%(2.0*np.pi)-np.pi)>np.pi)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    LS = np.sum(dW**2.0)
    LS+=P
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LS
    if verbose:
        return(p0,LS)
    return(LS)

def leastSq2_nohr(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(x.shape[0]-1)*dt,x.shape[0])
    z,g0,lda,mu,nu,chi,rho,asym,dasym = p0
    ix = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,g0,lda,mu,nu,chi,rho,asym,dasym).transpose(), x[0].transpose(), t)
    dW = (x-ix+np.pi)%(2.0*np.pi)-np.pi
    # Penalty for large changes
    P = np.sum(((ix[:-1]-ix[1:]+np.pi)%(2.0*np.pi)-np.pi)>np.pi)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    LS = np.sum(dW**2.0)
    LS+=P
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LS
    if verbose:
        return(p0,LS)
    return(LS)

def leastSq2_noasymhr(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(x.shape[0]-1)*dt,x.shape[0])
    z,g0,lda,mu,nu,chi,rho = p0
    ix = odeint(lambda nx,ta: -ndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,g0,lda,mu,nu,chi,rho).transpose(), x[0].transpose(), t)
    dW = (x-ix+np.pi)%(2.0*np.pi)-np.pi
    # Penalty for large changes
    P = np.sum(((ix[:-1]-ix[1:]+np.pi)%(2.0*np.pi)-np.pi)>np.pi)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    LS = np.sum(dW**2.0)
    LS+=P
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LS
    if verbose:
        return(p0,LS)
    return(LS)

# This function handles revised potential: derived by S. Wilshin Oct. 2012
# Starting without asymetry or hystereses: AJS 07-12-2012
def leastSq2_r2(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    dt = (sqrtdt**2.0)
    t = np.linspace(0,(x.shape[0]-1)*dt,x.shape[0])
    z,lda,mu,nu,chi,rho = p0
    ix = odeint(lambda nx,ta: -nndV(nx.transpose(),v[min(int(np.floor(ta/dt)),x.shape[0]-1)],z,lda,mu,nu,chi,rho).transpose(), x[0].transpose(), t)
    dW = (x-ix+np.pi)%(2.0*np.pi)-np.pi
    # Penalty for large changes
    P = np.sum(((ix[:-1]-ix[1:]+np.pi)%(2.0*np.pi)-np.pi)>np.pi)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    LS = np.sum(dW**2.0)
    LS+=P
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LS
    if verbose:
        return(p0,LS)
    return(LS)

efac = np.array([1,0,-1])
def dV(x,a,epsilon,sigma,eta,gamma,b,vel,mu,nu,chi,rho):
    u = (-2.0*a*np.sin(2.0*x)+0.5*epsilon*efac*np.cos(x))*(mu+vel*nu)
    u[0] += sigma*np.cos(x[0])*np.sin(x[2])*(mu+vel*nu)
    u[1] += -(eta*np.sin(x[1])+gamma*np.sin(x[1]))*(mu+vel*nu)
    u[2] += sigma*np.cos(x[2])*np.sin(x[0])*(mu+vel*nu)
    u += -b*np.sin(x)*(chi+vel*rho)
    return(u)

def dVBulk(x,a,epsilon,sigma,eta,gamma,b,vel,mu,nu,chi,rho):
    u = (((-2.0*a*np.sin(2.0*x)+0.5*epsilon*efac*np.cos(x)).transpose())*(mu+vel*nu)).transpose()
    u[:,0] += ((sigma*np.cos(x[:,0])*np.sin(x[:,2])).transpose()*(mu+vel*nu)).transpose()
    u[:,1] += ((-(eta*np.sin(x[:,1])+gamma*np.sin(x[:,1]))).transpose()*(mu+vel*nu)).transpose()
    u[:,2] += ((sigma*np.cos(x[:,2])*np.sin(x[:,0])).transpose()*(mu+vel*nu)).transpose()
    u += ((-b*np.sin(x)).transpose()*(chi+vel*rho)).transpose()
    return(u)

def makea(sigma,a0ma2):
  return(np.array([
                   (a0ma2+(sigma-1.0)/2.0)/2.0,
                   (sigma-1.0)/2.0,                   
                   (-a0ma2+(sigma-1.0)/2.0)/2.0
                  ]))

def makeb(lda,b0mb2):
  return(np.array([
                   (b0mb2-lda)/2.0,
                   1.0-lda,                   
                   -(b0mb2+lda)/2.0
                  ]))
       
def simulate(p0,x0,v,sqrtdt, S,da=0.0,db=0.0):
    x = np.zeros((v.size,3))
    lda = p0[0]
    sigma = p0[1]
    mu = p0[2]
    nu = p0[3]
    chi = p0[4]
    rho = p0[5]
    '''gamma = p0[2]
    mu = p0[3]
    nu = p0[4]
    chi = p0[5]
    rho = p0[6]'''
    gamma = 1.0
    epsilon = 1.0
    eta = -sigma
    a = makea(sigma,da)
    b = makeb(lda,db)
    x[0] = x0
    for i in range(v.size-1):
        x[i+1] = x[i]-dV(x[i],a,epsilon,sigma,eta,gamma,b,v[i],mu,nu,chi,rho)*(sqrtdt**2.0)+S*sqrtdt*np.dot(U.transpose(),(np.sqrt(D)*np.random.randn(3)))#np.sum(S*dt*np.dot(g3T,random.gauss(0,1)),1)
    return(x)

       
def logLike(p0,x,v,sqrtdt,verbose=False,talking=True,da=0.0,db=0.0):
    lda = p0[0]
    sigma = p0[1]
    mu = p0[2]
    nu = p0[3]
    chi = p0[4]
    rho = p0[5]
    '''gamma = p0[2]
    mu = p0[3]
    nu = p0[4]
    chi = p0[5]
    rho = p0[6]'''
    gamma = 1.0
    epsilon = 1.0
    eta = -sigma
    a = makea(sigma,da)
    b = makeb(lda,db)
    dW = ((x[1:]-x[:-1]+np.pi)%(2.0*np.pi)-np.pi+(sqrtdt**2.0)*dVBulk(x,a,epsilon,sigma,eta,gamma,b,v,mu,nu,chi,rho)[:-1,:])/(sqrtdt)
    dW = np.dot(np.dot(np.diag(1/np.sqrt(D)),U),dW.transpose()).transpose()
    S2 = np.var(dW)
    dP = (1.0/np.sqrt(2.0 * np.pi*S2))*np.exp(-(dW**2.0)/(2.0*S2))
    LL = np.sum(-np.log(dP))
    if verbose:
        pl.figure()
        pl.hist(dW/np.sqrt(S2),100)
        pl.title("Histogram of rescaled and transformed residuals, should be gaussian")
    if talking:
        print p0, LL
    if verbose:
        return(p0,np.sqrt(S2),LL)
    return(LL)
    
def V(x,vel,z,g0,lda,mu,nu,chi,rho,asym,dasym,hr=0.0):
    '''
    To aid in understanding, this function returns the potential evaluated at x
    Start with this, then do a bulk.
    
    gamma = np.array([g,z-1.0,g])
    b = np.array([
                   -lda/2.0,
                   1.0-lda,                   
                   -lda/2.0
                  ])
    # Walk potential              
    u = np.array([gamma*cos(x) + z/4.0*(sin(x[1])*sin(x[0]) + sin(x[2])*sin(x[1])) + 0.50*(1.0-z)*sin(x[2])*sin(x[0]) + 0.25*(1.0+2.0*g)*cos(x[1])*cos(x[0]) + 0.50*(sin(x[2])*cos(x[1]) + sin(x[2])*cos(x[0]) - sin(x[0])*cos(x[1]) - sin(x[0])*cos(x[2])))
    # Multiply walk by a scale factor plus velocity * another scale factor
    u *= (mu+vel*nu)
    # Add in the trot potential
    u += b*cos(x)*(chi+vel*rho)
    # ? We hit the rel leg phases with the matrix for computing the induced metric
    # This gives us new numbers in ph, slightly different. Is this so that 
    # the hyteresis we add is in correct scale? Don't get it.
    ph = np.dot(toroidalGeometry.g3T,x)
    # This sum makes no sense -- the difference here reduces it to a scalar. Error?
    # Should it also be twice the spatial frequency? Does ph introduce freq?
    u[0]-=hr*2*np.sum(2.0*(ph[0]-ph[2]))
    u[2]+=hr*2*np.sum(2.0*(ph[0]-ph[2]))
    # Here we simply add a line to x0 and subtract it from x2...
    # But this is in the gradient of V... so in V this looks like?
    # It would be linear in x[0] and x[2]:
    u[0]+=(asym+dasym*vel)*x[0] # Grad V would then strip off x[0]... etc.
    '''
    u[2]-=(asym-dasym*vel)*x[2] # This is not a trig function -> problem?
    return(u)
