'''
Created on 4 Feb 2011

@author: swilshin
'''

import numpy as np  

from splineInterpolationLimit import splineInterpolationLimit     
from quadratics import findQuadratic
from quadratics import findQuadraticPeak
  
def eliminatePeaks(opeaks,maxPeakRangeFraction,maxVariance):
    peaks=opeaks[:]
    converged=True
    peakLocations,peakValues=zip(*peaks)
    peakInverseQualityMetric=np.abs(peakValues-np.mean(peakValues))/np.std(peakValues)        
    peakFraction=(np.max(peakValues)-np.min(peakValues))/np.max(peakValues)
    if (peakFraction<maxPeakRangeFraction):
        return (peaks,True)
    inverseQualities=zip(peakLocations,peakInverseQualityMetric)
    badPeak,badPeakInverseQuality=sorted(inverseQualities,key=lambda s: s[1],reverse=True)[0]
    if badPeakInverseQuality>maxVariance:
        peaks=[(k,v) for k,v in peaks if not k==badPeak]
        converged=False
    return((peaks,converged))
            
def findPeaks(x,maxX,maxPeakRangeFraction,maxVariance):
        peak=np.max(x)
        peakLocations=[i for i in range(1,len(x)-1) if x[i]>x[i-1] and x[i]>x[i+1] and x[i]>(peak*maxX)]
        # Now remove remaining false peaks
        peakLocations[:]
        peaks=[x[k] for k in peakLocations]
        converged=False
        peaks=zip(peakLocations,peaks)
        while not converged:
            peaks,converged=eliminatePeaks(peaks,maxPeakRangeFraction,maxVariance)
        return peaks  

def simplePeakFind(d,minimumHeight):
    nD=d[1:]-d[:-1]
    return([k+2 
            for k in range(0,len(nD)-1) 
            if d[k]>minimumHeight and 
              (
               (nD[k]==0 and k!=0 and nD[k-1]!=0) or 
               (nD[k]!=0 and nD[k+1]!=0 and (nD[k]/abs(nD[k])-nD[k+1]/abs(nD[k+1]))==2)
              )
          ])

def getPeak(signal,sampleRate):
        if len(signal)==0:
            return(0.0) # An empty signal has no power at any frequency
        peakFrequency=signal.argmax()
        if len(signal)<splineInterpolationLimit: # If signal is this short then quadratic interpolation wont work
            return(peakFrequency/(2.0*len(signal))*sampleRate)
        weights=signal[(peakFrequency-1):(peakFrequency+2)]
        freqs=np.array([-1,0,1])+peakFrequency
        quadratic=findQuadratic(freqs,weights)
        return(findQuadraticPeak(quadratic)/(2.0*len(signal))*sampleRate)
