'''
Created on 4 Feb 2011

@author: swilshin
'''

from cmath import phase
from scipy.optimize import fmin
from scipy.signal import hilbert
import numpy as np

from quadratics import quadraticChiSquared

def angularCentralTendancy(x,threshold=0.001,maxIt=100,ctf=np.mean,t_ctf=None,ctfIsSpread=False):
    if t_ctf==None:
        t_ctf=ctf
    t_x = np.array(x)%(2.0*np.pi)
    ch = [t_ctf(t_x>n*np.pi/2.0) and t_ctf(t_x<n+1*np.pi/2.0) for n in range(4)] 
    n = np.array(ch).argmax()
    t_av = t_ctf((t_x-n*np.pi/2.0+np.pi)%(2.0*np.pi))+n*np.pi/2.0-np.pi
    # Now iterate till convergence
    ot_av=0.0
    it=0
    while it==0 or (np.abs(ot_av-t_av)>threshold and it<maxIt):
        ot_av=t_av
        t_av = t_ctf((t_x+np.pi-ot_av)%(2.0*np.pi))-np.pi+ot_av
        it+=1
    if ctfIsSpread:
        t_av = ctf((t_x+np.pi-ot_av)%(2.0*np.pi))
    else:
        t_av = ctf((t_x+np.pi-ot_av)%(2.0*np.pi))-np.pi+ot_av
        t_av = t_av%(2.0*np.pi)
    return(t_av)

def angularCovariance(x,y,threshold=0.001,maxIt=100,t_ctf=np.mean):
    '''
    Compute the covariance matrix of two angular variables
    '''
    t_x = np.array(x)%(2.0*np.pi)
    ch = [t_ctf(t_x>n*np.pi/2.0) and t_ctf(t_x<n+1*np.pi/2.0) for n in range(4)] 
    n = np.array(ch).argmax()
    tx_av = t_ctf((t_x-n*np.pi/2.0+np.pi)%(2.0*np.pi))+n*np.pi/2.0-np.pi
    
    t_y = np.array(y)%(2.0*np.pi)
    ch = [t_ctf(t_y>n*np.pi/2.0) and t_ctf(t_y<n+1*np.pi/2.0) for n in range(4)] 
    n = np.array(ch).argmax()
    ty_av = t_ctf((t_y-n*np.pi/2.0+np.pi)%(2.0*np.pi))+n*np.pi/2.0-np.pi

    
    # Now iterate till convergence
    otx_av=0.0
    oty_av=0.0
    it=0
    while it==0 or (np.abs(otx_av-tx_av)>threshold and np.abs(oty_av-ty_av)>threshold and it<maxIt):
        otx_av=tx_av
        oty_av=ty_av
        tx_av = t_ctf((t_x+np.pi-otx_av)%(2.0*np.pi))-np.pi+otx_av
        ty_av = t_ctf((t_y+np.pi-oty_av)%(2.0*np.pi))-np.pi+oty_av
        it+=1
        
    t_av = np.cov((t_x+np.pi-otx_av)%(2.0*np.pi),(t_y+np.pi-oty_av)%(2.0*np.pi))[1][0]
    return(t_av)

def makePhase(x):
    return(np.array([phase(p) for p in hilbert(x)]))
        
def deMod(phi,tol=3.0/2.0):
    count = 0
    nphi = []
    for p,q in zip(phi[:-1],phi[1:]):
        nphi.append(p+count*2*np.pi)
        if p-q>tol*np.pi:
            count = count+1
        if p-q<-tol*np.pi:
            count = count-1            
    nphi.append(phi[-1]+count*2*np.pi)
    return(np.array(nphi))

def polyComplex(phis,tau):
    polys=[]
    for phi in phis:
        phi = deMod(phi)
        gr=(tau[-1]-tau[0])/(phi[-1]-phi[0])
        t0=tau[-1]-gr*phi[0]
        a=[t0,gr,0]
        a=fmin(quadraticChiSquared,a,args=tuple([tau,phi]))
        polys.append(a)
    a=[]
    b=[]
    c=[]
    d=[]
    for p in polys:
        na=[]
        nb=[]
        nc=[]
        nd=[]
        for q in polys:
            na.append(np.sqrt(p[1]/q[1]))
            nb.append((q[0]-p[0])/np.sqrt(p[1]*q[1]))
            nc.append(q[2]/np.sqrt(p[1]*q[1]))
            nd.append(-p[2]/np.sqrt(p[1]*q[1]))
            #print np.sqrt(p[1]/q[1]), np.sqrt(q[1]/p[1]),(q[0]-p[0])/np.sqrt(p[1]*q[1]),q[2]/np.sqrt(p[1]*q[1]),-p[2]/np.sqrt(p[1]*q[1])
        a.append(na)
        b.append(nb)
        c.append(nc)
        d.append(nd)
    return(tuple([a,b,c,d]))

