
import numpy as np
import pylab as pl

def cdot(x,y):
    return((x.conjugate()*y).real)

def reCenterCl(phaseFun):
    def newPhase(s,phi,l=1000.0,r=0.4):
        # Find a good initial center
        c = np.zeros_like(phi)
        c[0:2] = 0.0#np.mean(phi)
        for i in range(2,c.size):
            xr0 = phi[i-2]-phi[i-1]
            xr1 = phi[i-1]-phi[i]
            xs0 = phi[i-2]+phi[i-1]
            xs1 = phi[i-1]+phi[i]
            M = np.array([
                          [cdot(xr0,xr0), cdot(xr0,xr1)],
                          [cdot(xr0,xr1), cdot(xr1,xr1)]
                        ])
            k = (1.0/2.0)*np.array([cdot(xr0,xs0), cdot(xr1,xs1)])
            if np.linalg.det(M)!=0.0:
                alpha = np.dot(np.linalg.inv(M),k)
                dc = xr0*alpha[0]+xr1*alpha[1]-c[i-1]
            else:
                dc=0.0
            c[i] = c[i-1] + r*dc*np.exp(-np.sqrt(cdot(dc,dc))/l)*(np.exp(1.0)/l)
        # Compute the shifts one should apply to the center
        # Return the new value for the analytic signal
        return(phi-c)
        
    def wrap(s,x):
        phi = phaseFun(s,x)
        nphi = list()
        for p in phi:
            nphi.append(newPhase(s,p))
            z = p
            zp = nphi[-1]
            '''pl.figure()
            pl.plot(zp.real,zp.imag,'r-')
            pl.plot(z.real,z.imag,'b-')
            pl.figure()
            pl.plot(np.angle(zp),'r-')
            pl.plot(np.angle(z),'b-')
            pl.show()'''
        return(np.array(nphi))
    
    return(wrap)

def reCenter(phaseFun):
    def newPhase(x,l=1.0,r=0.01):
        phi = phaseFun(x)
        # Find a good initial center
        c = np.zeros_like(phi)
        c[0:2] = np.mean(phi)
        for i in range(2,c.size):
            xr0 = x[i-2]-x[i-1]
            xr1 = x[i-1]-x[i]
            xs0 = x[i-2]+x[i-1]
            xs1 = x[i-1]+x[i]
            M = np.array([
                          [cdot(xr0,xr0), cdot(xr0,xr1)],
                          [cdot(xr0,xr1), cdot(xr1,xr1)]
                        ])
            k = (1.0/2.0)*np.array([cdot(xr0,xs0), cdot(xr1,xs1)])
            alpha = np.dot(np.linalg.inv(M),k)
            dc = xr0*alpha[0]+xr1*alpha[1]-c[i-1]
            c[i] = c[i-1] + r*dc*np.exp(-np.sqrt(cdot(dc,dc))/l)*(np.exp(1.0)/l)
        # Compute the shifts one should apply to the center
        # Return the new value for the analytic signal
        return(phi-c)
    return(newPhase)

@reCenter
def phaseRepeater(x):
    return(x)

if __name__=="__main__":
    import pylab as pl
    t = np.linspace(0,300,30000)
    a = np.sin(t)+0.1*np.sin(2.0*t)+10.0*np.sin(t/30.0)
    a[10000:20000] += 0.5
    b = np.cos(t)
    z = a + b*1.0j
    zp = phaseRepeater(z)
    pl.figure()
    pl.plot(zp.real,zp.imag,'r-')
    pl.plot(z.real,z.imag,'b-')
    pl.figure()
    pl.plot(np.angle(z))
    pl.plot(np.angle(zp))
    pl.show()