'''
Created on 4 Feb 2011

@author: swilshin
'''

import numpy as np


class NotAQuadratic(Exception):
    '''
    When it is not possible to find an anchor point for the 
    foot markers (shoulder, pelvis, etc) this represents a serious 
    problem this module cannot handle and an exception should be 
    thrown, this exception.
    '''
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def findQuadraticPeak(A):
    if A[0]==0:
        raise NotAQuadratic("Looking for the peak of a order one polynomial is foolhardy.")
    return(-A[1]/(2*A[0]))

def findQuadratic(X,Y):
    A = (
         ((Y[1] - Y[0]) * (X[0] - X[2]) + 
         (Y[2] - Y[0]) * (X[1] - X[0])) / 
        ((X[0] - X[2]) * (X[1] ** 2 - X[0] ** 2) + 
         (X[1] - X[0]) * (X[2] ** 2 - X[0] ** 2))
        )
    B = (
         ((Y[1] - Y[0]) - A * (X[1] ** 2 - X[0] ** 2)) 
        / (X[1] - X[0])
        )
    C = Y[0] - A * X[0] ** 2 - B * X[0]
    return([A,B,C])

def solveQuadratic(A):
    princ = -A[1]
    delta = A[1]**2.0-4*A[0]*A[2]
    return([ (princ+np.sqrt(delta))/(2.0*A[0]), (princ-np.sqrt(delta))/(2.0*A[0]) ])

def quadraticChiSquared(a,y,x):
    yE=np.array(a[0]+a[1]*x+a[2]*(x**2.0))
    return(np.sum((yE-y)**2))


# Testing
def testFindQuadraticPeak():
    print "Testing findQuadraticPeak"
    t1 = findQuadraticPeak([1,-2,0])==1
    print "Test 0001 Passed: ", t1
    t2 = findQuadraticPeak([1,-18,0])==9
    print "Test 0002 Passed: ", t2
    testsPassed = t1 and t2
    return(testsPassed)

def testFindQuadratic():
    print "Testing findQuadratic"
    q = findQuadratic([0,1,2],[1,2,3])
    t1 = q[0]==0.0
    print "Test 0001 Passed: ", t1
    t2 = q[1]==1.0
    print "Test 0002 Passed: ", t2
    t3 = q[2]==1.0
    print "Test 0003 Passed: ", t3
    testsPassed = t1 and t2 and t3
    return(testsPassed)

def testSolveQuadratic():
    print "Testing solveQuadratic"
    s = solveQuadratic([1,0,-1])
    t1 = s[0]==1
    print "Test 0001 Passed: ", t1
    t2 = s[1]==-1
    print "Test 0002 Passed: ", t2
    s = solveQuadratic([1,-2,1])
    t3 = s[0]==1
    print "Test 0003 Passed: ", t3
    t4 = s[1]==1
    print "Test 0004 Passed: ", t4
    testsPassed = t1 and t2 and t3 and t4
    return(testsPassed)

def testQuadraticChiSquared():
    print "Testing quadraticChiSquared"
    t1 = quadraticChiSquared(np.array([1,1,0]),np.array([1,2]),np.array([0,1]))==0
    print "Test 0001 Passed: ", t1
    t2 = quadraticChiSquared(np.array([1,1,0]),np.array([1,1]),np.array([0,1]))==1
    print "Test 0002 Passed: ", t2
    testsPassed = t1 and t2
    return(testsPassed)

def testQuadratics():
    print "Testing quadratics module, True indicates test passed."        
    t1 = testFindQuadraticPeak()
    t2 = testFindQuadratic()
    t3 = testSolveQuadratic()
    t4 = testQuadraticChiSquared()
    testsPassed = t1 and t2 and t3 and t4
    if testsPassed:
        print "All tests PASSED in module quadratics."
    else:
        print "Some tests FAILED in module quadratics!"
    return(testsPassed)

if __name__=='__main__':
    '''
    Run some tests, should print True if everything worked
    '''
    testQuadratics()