'''
Created on 11 Jun 2011

@author: swilshin
'''

import numpy as np
from scipy.optimize import fmin_l_bfgs_b
#from enthought.mayavi import mlab


from toroidalGeometry import g3TP
from toroidalGeometry import g3T

defaultRad = 0.01

def stance4StabScores(form):
    stance4 = {
               (0,0,0,0):-0.3,
               (0,0,0,1):-0.2, (0,0,1,0):-0.2, (0,1,0,0):-0.2, (1,0,0,0):-0.2,
               (0,0,1,1): 0.0, (1,1,0,0): 0.0, (1,0,1,0):-0.1, (0,1,0,1):-0.1, (1,0,0,1):-0.15, (0,1,1,0):-0.15,
               (0,1,1,1): 0.0, (1,0,1,1): 0.0, (1,1,0,1): 0.0, (1,1,1,0): 0.0,
               (1,1,1,1): 0.1
              }
    return(-(stance4[form]-0.1))

def phiFromThetaPsi(psi, theta):
    phi = np.zeros(4)
    phi[1:] = np.cumsum(theta)
    phi += psi
    phi %= (2.0*np.pi)
    return(phi)
    
def stance4StabScoreWrapper(psi,theta,phdf):
    phi = phiFromThetaPsi(psi, theta)
    return(stance4StabScores(tuple(phi<phdf)))

def getEndOfRegion(startOfRegion,theta,phdf):
    phi = phiFromThetaPsi(startOfRegion,theta)
    distanceToWrap = (2.0*np.pi)-phi
    distanceToLift = (phdf-phi)%(2.0*np.pi)
    distanceToWrap += 2.0*np.pi*(distanceToWrap==0)
    distanceToLift += 2.0*np.pi*(distanceToLift==0)
    mW = np.min(distanceToWrap)
    mL = np.min(distanceToLift)
    mT = min(mW,mL)
    return(startOfRegion+mT)

def computeStabScore(theta,df=0.5):
    phdf = 2.0*np.pi*df
    startOfRegion=0
    stabScore=0
    complete = False
    while not complete:
        endOfRegion = getEndOfRegion(startOfRegion,theta,phdf)
        if np.isnan(endOfRegion):
            stabScore = np.nan
            complete = True
        if endOfRegion>2.0*np.pi:
            endOfRegion = 2.0*np.pi
            complete = True
        stabScore += (endOfRegion-startOfRegion)*stance4StabScoreWrapper((startOfRegion+endOfRegion)/2.0,theta,phdf)
        startOfRegion = endOfRegion
    stabScore /= 2.0*np.pi
    return(stabScore)

def theta3AtTheta12(t,rad=defaultRad):
    return(np.dot(np.linalg.inv(g3TP),rad*np.array([np.sin(t[0])*np.cos(t[1]),np.sin(t[0])*np.sin(t[1]),np.cos(t[1])])))

def computeStabScoreAtRad(t,theta0,df,rad=defaultRad,givedZ=False):
    dt = theta3AtTheta12(t,rad)
    val = computeStabScore(theta0+dt,df)
    if givedZ:
        return(val,dt)
    else:
        return(val)

def stabMinAtRad(theta0,df,rad=defaultRad):
    # Cycle over angles and find the direction of maximal stability score
    am = 0
    bm = 0
    minSS = computeStabScoreAtRad([am,bm],theta0,df,rad)
    for a in np.linspace(0,np.pi,10):
        for b in np.linspace(0,2.0*np.pi,10):
            nSS = computeStabScoreAtRad([a,b],theta0,df,rad)
            if nSS<minSS:
                minSS = nSS
                am = a
                bm = b
                
    return(fmin_l_bfgs_b(computeStabScoreAtRad,[am,bm],args=(theta0,df,rad),bounds=[[0,np.pi],[0,np.pi]],approx_grad=True,factr=10))

'''
def makeStabilityPlot(df=0.75):
    N=4
    #x, y, z = np.pi+np.mgrid[-N:(N+1), -N:(N+1), -N:(N+1)]*(np.pi/N)
    x, y, z = np.pi+np.mgrid[-N:1, -N/2:(N/2+1), -0:(N+1)]*(np.pi/N)
    N=N/2
    u = np.zeros_like(x)
    v = np.zeros_like(x)
    w = np.zeros_like(x)
    for idx in range(2*N+1):
        for idy in range(2*N+1):
            for idz in range(2*N+1):
                t = np.array([x[idx,idy,idz],y[idx,idy,idz],z[idx,idy,idz]])
                th = stabMinAtRad(t,df)
                me,dt = computeStabScoreAtRad(th[0],t,df,givedZ=True)
                ms = computeStabScore(t,df)                
                dt = dt / np.sqrt(np.dot(dt.transpose(),np.dot(g3T,dt)))
                dt *= 0.4-ms
                u[idx,idy,idz] = dt[0]
                v[idx,idy,idz] = dt[1]
                w[idx,idy,idz] = dt[2]
    scale = np.max(np.array([u,v,w]))
    print x,y,z,y,v,w
    obj = mlab.flow(x, y, z, u, v, w, line_width=2)
    mlab.show()
    return obj"""    

"""if __name__=="__main__":
    theta01 = np.array([np.pi/2.0,np.pi,-np.pi/2])
    df1=0.75
    rad1=0.01
    t1 = stabMinAtRad(theta01,df1,rad1)
    theta02 = np.array([3.27,5.70,4.99])
    df2=0.25
    rad2=0.01
    t2 = stabMinAtRad(theta02,df2,rad2)
    print "Walk"
    print computeStabScoreAtRad(t1[0],theta01,df1,rad1,True)
    print "Gallop"        
    print computeStabScoreAtRad(t2[0],theta02,df2,rad2,True)
'''