'''
Created on 24 Apr 2011

@author: swilshin
'''

import numpy as np
from scipy.linalg import eig

# Note: In this convention the first component of the four vector results from dot
# of to3Torus with a four vector of leg phases is the global phase.
to3Torus = np.array([
                     [  1,  1,  1,  1 ],
                     [  1, -1,  0,  0 ],
                     [  0,  1, -1,  0 ],
                     [  0,  0,  1, -1 ]
                   ])

# Note: in going back to 4 torus, the first element of the four-vector is the global
# phase, the latter three are the leg phase differences.
to4Torus = 0.25*np.array([
                          [  1,  3,  2,  1 ],
                          [  1, -1,  2,  1 ],
                          [  1, -1, -2,  1 ],
                          [  1, -1, -2, -3 ]                       
                        ])

g3T = np.array([
                   [3.0/4.0, 1.0/2.0, 1.0/4.0],
                   [1.0/2.0, 1.0,     1.0/2.0],
                   [1.0/4.0, 1.0/2.0, 3.0/4.0],
                  ])

'''
The following was derived using:
import sympy
p0,p1,p2,p3,p4,t0,t1,t2,t3,t4,t5,psi = sympy.symbols('p0 p1 p2 p3 p4 t0 t1 t2 t3 t4 t5 psi')
eqn = sympy.solvers.solve([p0-t0+t1,p1-t1+t2,p2-t2+t3,p3-t3+t4,p4-t4+t5,t0+t1+t2+t3+t4+t5-psi],t0,t1,t2,t3,t4,t5)
g5T = np.zeros((5,5))
eta = [p0,p1,p2,p3,p4]
for j in range(len(eta)):
    for k in range(len(eta)):
        g5T[j,k] = np.sum([sympy.diff(eqn[th],eta[j])*sympy.diff(eqn[th],eta[k]) for th in eqn])
'''
g5T = np.array([[ 0.83333333,  0.66666667,  0.5       ,  0.33333333,  0.16666667],
       [ 0.66666667,  1.33333333,  1.        ,  0.66666667,  0.33333333],
       [ 0.5       ,  1.        ,  1.5       ,  1.        ,  0.5       ],
       [ 0.33333333,  0.66666667,  1.        ,  1.33333333,  0.66666667],
       [ 0.16666667,  0.33333333,  0.5       ,  0.66666667,  0.83333333]])
       
g7T = np.array([[ 0.875,  0.75 ,  0.625,  0.5  ,  0.375,  0.25 ,  0.125],
       [ 0.75 ,  1.5  ,  1.25 ,  1.   ,  0.75 ,  0.5  ,  0.25 ],
       [ 0.625,  1.25 ,  1.875,  1.5  ,  1.125,  0.75 ,  0.375],
       [ 0.5  ,  1.   ,  1.5  ,  2.   ,  1.5  ,  1.   ,  0.5  ],
       [ 0.375,  0.75 ,  1.125,  1.5  ,  1.875,  1.25 ,  0.625],
       [ 0.25 ,  0.5  ,  0.75 ,  1.   ,  1.25 ,  1.5  ,  0.75 ],
       [ 0.125,  0.25 ,  0.375,  0.5  ,  0.625,  0.75 ,  0.875]])
       
g3Td,U = eig(g3T)
U = U.real
g3TD = np.diag(g3Td).real
g3TP = np.dot(np.sqrt(g3TD),U.transpose())
g3TPInv = np.linalg.inv(g3TP)

def geodesicDistance3T(x0,x1):
    delta = ((x0-x1)+np.pi)%(2.0*np.pi)-np.pi
    return(np.mean(delta*np.dot(g3T,delta)))

def geodesicDistance5T(x0,x1):
    delta = ((x0-x1)+np.pi)%(2.0*np.pi)-np.pi
    return(np.mean(delta*np.dot(g5T,delta)))

def geodesicDistance7T(x0,x1):
    delta = ((x0-x1)+np.pi)%(2.0*np.pi)-np.pi
    return(np.mean(delta*np.dot(g7T,delta)))

def geodesicDirection(x0,x1):
    t_x0 = np.zeros(4)
    t_x0[1:] = x0%(2.0*np.pi)
    t_x0 = np.dot(to4Torus,t_x0)
    t_x0 %= (2.0*np.pi)
    t_x1 = np.ones(4)
    t_x1[1:] = x1%(2.0*np.pi)
    t_x1 = np.dot(to4Torus,t_x1)
    t_x1 %= (2.0*np.pi)
    v = (((t_x1-t_x0)+np.pi)%(2.0*np.pi))-np.pi
    v3 = ((np.dot(to3Torus,v)+np.pi)%(2.0*np.pi))-np.pi
    return(v3[1:])

def geodesicCurve(psi,x0,x1):
    '''
    Compute the value of a geodesic starting with psi (sum of the 4 phases) at zero and ending at psi is unity with psi at value psi. 
    '''
    t_x0 = np.zeros(4)
    t_x0[1:] = x0%(2.0*np.pi)
    t_x0 = np.dot(to4Torus,t_x0)
    t_x0 %= (2.0*np.pi)
    t_x1 = np.ones(4)
    t_x1[1:] = x1%(2.0*np.pi)
    t_x1 = np.dot(to4Torus,t_x1)
    t_x1 %= (2.0*np.pi)
    v = (((t_x1-t_x0)+np.pi)%(2.0*np.pi))-np.pi
    x4 = ((t_x0+psi*v+np.pi)%(2.0*np.pi))-np.pi
    # Now modulo to ensure we are on the shortest segment
    # Project back onto 3 torus and return
    x3 = ((np.dot(to3Torus,x4)+np.pi)%(2.0*np.pi))-np.pi
    return(x3[1:])